"""Status panel widget for displaying current app state."""

from rich.text import Text
from textual.reactive import reactive
from textual.timer import Timer
from textual.widgets import Static

from voxing.state import AppState
from voxing.themes import (
    FOREGROUND,
    PALETTE_1,
    PALETTE_3,
    PALETTE_4,
    PALETTE_5,
    PALETTE_6,
)


class StatusPanel(Static):
    """Widget for displaying current status and state."""

    SPINNER_FRAMES = ["⠋", "⠙", "⠹", "⠸", "⠼", "⠴", "⠦", "⠧", "⠇", "⠏"]

    current_state: reactive[AppState] = reactive(AppState.LOADING)
    tts_enabled: reactive[bool] = reactive(False)
    status_message: reactive[str | None] = reactive(None)
    _frame_index: int = 0
    _animation_timer: Timer | None = None
    _ephemeral_timer: Timer | None = None

    def show_status_message(self, message: str) -> None:
        """Show a status message that persists until explicitly cleared."""
        if self._ephemeral_timer is not None:
            self._ephemeral_timer.stop()
            self._ephemeral_timer = None
        self.status_message = message

    def show_ephemeral_message(self, message: str, timeout: float = 3.0) -> None:
        """Show a temporary status message that auto-clears."""
        if self._ephemeral_timer is not None:
            self._ephemeral_timer.stop()
        self.status_message = message
        self._ephemeral_timer = self.set_timer(timeout, self._clear_ephemeral)

    def clear_status_message(self) -> None:
        """Clear any active status message."""
        self.status_message = None
        if self._ephemeral_timer is not None:
            self._ephemeral_timer.stop()
            self._ephemeral_timer = None

    def _clear_ephemeral(self) -> None:
        """Clear the ephemeral status message."""
        self.status_message = None
        self._ephemeral_timer = None

    def watch_status_message(self) -> None:
        """Called when status_message changes."""
        self.update_display()

    def watch_current_state(self, new_state: AppState) -> None:
        """Called when current_state changes."""
        if self._animation_timer is not None:
            if new_state == AppState.READY:
                self._animation_timer.pause()
            else:
                self._animation_timer.resume()
        self.update_display()

    def watch_tts_enabled(self, new_value: bool) -> None:
        """Called when tts_enabled changes."""
        self.update_display()

    def update_display(self) -> None:
        """Update the panel display with current state."""
        content = Text()

        content.append("Voxing", style=PALETTE_4)
        content.append("  •  ")

        if self.current_state != AppState.READY:
            spinner = self.SPINNER_FRAMES[self._frame_index]
            content.append(f"{spinner} ", style=self._get_state_color())
        content.append(str(self.current_state), style=self._get_state_color())

        if self.status_message:
            content.append(f"  •  {self.status_message}", style=f"dim {PALETTE_3}")

        if self.tts_enabled:
            content.append("  •  TTS", style=PALETTE_6)

        self.update(content)

    def _advance_animation(self) -> None:
        """Advance to the next animation frame."""
        self._frame_index = (self._frame_index + 1) % len(self.SPINNER_FRAMES)
        self.update_display()

    def _get_state_color(self) -> str:
        """Get the color for the current state."""
        color_map = {
            AppState.LOADING: PALETTE_3,  # #e0af68 (yellow) - processing
            AppState.READY: PALETTE_5,  # #bb9af7 (purple) - idle
            AppState.RECORDING: PALETTE_1,  # #f7768e (red) - active I/O
            AppState.TRANSCRIBING: PALETTE_3,  # #e0af68 (yellow) - processing
            AppState.THINKING: PALETTE_3,  # #e0af68 (yellow) - processing
            AppState.SYNTHESIZING: PALETTE_3,  # #e0af68 (yellow) - processing
            AppState.SPEAKING: PALETTE_1,  # #f7768e (red) - active I/O
        }
        return color_map.get(self.current_state, FOREGROUND)

    def on_mount(self) -> None:
        """Called when widget is mounted."""
        self.update_display()
        pause = self.current_state == AppState.READY
        self._animation_timer = self.set_interval(
            0.1, self._advance_animation, pause=pause
        )



"""Main Voxing TUI application."""

import asyncio
import contextlib
import time
from os.path import commonprefix

from rich.text import Text
from textual import on
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.containers import Container, Horizontal, Vertical
from textual.css.query import NoMatches
from textual.events import Key
from textual.reactive import reactive
from textual.timer import Timer
from textual.widgets import Footer, Label
from textual.worker import Worker

from voxing import audio as audio_mod
from voxing.config import settings
from voxing.models import Models, load_llm, load_stt, load_tts
from voxing.models import llm as llm_mod
from voxing.models import stt as stt_mod
from voxing.models import tts as tts_mod
from voxing.models.llm import ChatMessage
from voxing.state import AppState, InteractionStats
from voxing.themes import FOREGROUND, PALETTE_1, TOKYO_NIGHT
from voxing.widgets import (
    ChatInput,
    ConversationLog,
    MetricsPanel,
    ModelSelection,
    ModelSelector,
    StatusPanel,
)

COMMAND_DESCRIPTIONS: dict[str, str] = {
    "/record": "Record and process voice",
    "/transcribe": "Transcribe audio to text",
    "/tts": "Toggle text-to-speech",
    "/model": "Switch models",
    "/clear": "Clear conversation",
    "/exit": "Exit application",
}
COMMANDS = list(COMMAND_DESCRIPTIONS)


class VoxingApp(App):
    """Voxing Voice Assistant TUI Application."""

    CSS_PATH = "styles.tcss"
    TITLE = "Voxing Voice Assistant"

    BINDINGS = [
        Binding("q", "quit", "Quit", priority=True),
        Binding("ctrl+c", "clear_conversation", "Clear"),
    ]

    state: reactive[AppState] = reactive(AppState.LOADING)

    def __init__(self) -> None:
        """Initialize the Voxing TUI app."""
        super().__init__()
        self.models: Models = None
        self.is_processing = False
        self.tts_enabled = False
        self.chat_history: list[ChatMessage] = []
        self._active_worker: Worker[None] | None = None
        self._esc_pending: bool = False
        self._esc_timer: Timer | None = None

    def compose(self) -> ComposeResult:
        """Compose the app layout."""
        with Horizontal(id="status-bar"):
            yield StatusPanel(id="status-panel")
            yield MetricsPanel(id="metrics-panel")

        yield ConversationLog(id="conversation-log", wrap=True)

        with Vertical(id="bottom-section"), Container(id="input-container"):
            yield Label(id="command-hint", classes="hidden")
            yield ChatInput(id="user-input")

        yield Footer()

    def on_mount(self) -> None:
        """Called when app is mounted."""
        self.design = TOKYO_NIGHT

        text_area = self.query_one("#user-input", ChatInput)
        text_area.focus()
        text_area.show_line_numbers = False
        text_area.disabled = True

        status_panel = self.query_one("#status-panel", StatusPanel)
        status_panel.current_state = self.state
        status_panel.tts_enabled = self.tts_enabled

        self.run_worker(self._load_models())

    async def _load_models(self) -> None:
        """Load all models in the background."""
        loop = asyncio.get_running_loop()
        status_panel = self.query_one("#status-panel", StatusPanel)

        status_panel.show_status_message("Loading STT...")
        stt = await loop.run_in_executor(None, load_stt)

        status_panel.show_status_message("Loading LLM...")
        llm, tokenizer = await loop.run_in_executor(None, load_llm)

        status_panel.show_status_message("Loading TTS...")
        tts = await loop.run_in_executor(None, load_tts)

        status_panel.clear_status_message()

        self.models = Models(stt=stt, llm=llm, tts=tts, tokenizer=tokenizer)

        text_area = self.query_one("#user-input", ChatInput)
        text_area.disabled = False
        text_area.focus()
        self.state = AppState.READY

    def watch_state(self, new_state: AppState) -> None:
        """Called when state changes."""
        with contextlib.suppress(NoMatches):
            status_panel = self.query_one("#status-panel", StatusPanel)
            status_panel.current_state = new_state

    def on_key(self, event: Key) -> None:
        """Handle key events."""
        if event.key == "escape":
            if not self.is_processing:
                return
            event.prevent_default()
            event.stop()
            if self._esc_pending:
                self._cancel_processing()
            else:
                self._esc_pending = True
                self._show_status("Press ESC again to cancel", timeout=2.0)
                self._esc_timer = self.set_timer(2.0, self._reset_esc_pending)
            return

        if event.key != "tab":
            return

        text_area = self.query_one("#user-input", ChatInput)
        text = text_area.text.strip()

        if not text.startswith("/"):
            return

        matches = [cmd for cmd in COMMANDS if cmd.startswith(text.lower())]
        if not matches:
            return

        event.prevent_default()
        event.stop()

        replacement = matches[0] if len(matches) == 1 else commonprefix(matches)

        text_area.clear()
        text_area.insert(replacement)

    @on(ChatInput.Changed)
    def on_chat_input_changed(self, event: ChatInput.Changed) -> None:
        """Update command hints when text changes."""
        self._update_command_hints(event.text_area.text)

    @on(ChatInput.Submitted)
    def on_chat_input_submitted(self, event: ChatInput.Submitted) -> None:
        """Handle input submission."""
        if self.is_processing:
            return
        user_input = event.value.strip()
        if user_input:
            event.chat_input.clear()
            self._hide_command_hints()
            self._active_worker = self.run_worker(self._process_input(user_input))

    def _update_command_hints(self, text: str) -> None:
        """Update command hints based on current input."""
        hint_label = self.query_one("#command-hint", Label)
        typed = text.strip().lower()

        if typed.startswith("/"):
            matches = [cmd for cmd in COMMANDS if cmd.startswith(typed)]

            if matches:
                hint_text = Text()
                for i, cmd in enumerate(matches):
                    if i > 0:
                        hint_text.append("\n")
                    hint_text.append(cmd, style=f"bold {FOREGROUND}")
                    hint_text.append(f" {COMMAND_DESCRIPTIONS[cmd]}", style=FOREGROUND)
                hint_label.update(hint_text)
                hint_label.remove_class("hidden")
            else:
                self._hide_command_hints()
        else:
            self._hide_command_hints()

    def _hide_command_hints(self) -> None:
        """Hide command hints."""
        with contextlib.suppress(NoMatches):
            self.query_one("#command-hint", Label).add_class("hidden")

    async def _process_input(self, text: str) -> None:
        """Route user input to the appropriate handler."""
        if self.state == AppState.LOADING:
            return
        self.is_processing = True
        conv_log = self.query_one("#conversation-log", ConversationLog)

        try:
            if text.startswith("/"):
                command = text.lower().strip()
                if command == "/record":
                    await self._run_record_pipeline()
                elif command == "/transcribe":
                    await self._run_transcribe()
                elif command == "/tts":
                    self._toggle_tts()
                elif command == "/model":
                    self._open_model_selector()
                elif command == "/clear":
                    self.action_clear_conversation()
                elif command == "/exit":
                    self.exit()
                else:
                    conv_log.add_system_message(
                        f"Unknown command: {command}. Available: {', '.join(COMMANDS)}",
                        style=f"bold {PALETTE_1}",
                    )
            else:
                conv_log.add_user_message(text)
                await self._run_llm_pipeline(text)
        except Exception as e:
            conv_log.add_error(str(e))
            self.state = AppState.READY
        finally:
            self._active_worker = None
            self.is_processing = False

    async def _run_llm_pipeline(
        self, text: str, transcribe_time: float | None = None
    ) -> None:
        """Run LLM generation on text input."""
        conv_log = self.query_one("#conversation-log", ConversationLog)
        metrics_panel = self.query_one("#metrics-panel", MetricsPanel)

        self.state = AppState.THINKING

        llm_start = time.time()
        token_count = 0
        full_response = ""
        phrase_buffer = ""
        phrases: list[str] = []

        conv_log.start_streaming_response()

        async for token in llm_mod.generate_streaming(
            self.models.llm, self.models.tokenizer, text, history=self.chat_history
        ):
            full_response += token
            conv_log.update_streaming_response(token)
            token_count += 1

            if self.tts_enabled:
                phrase_buffer += token
                if "\n" in phrase_buffer:
                    phrase, phrase_buffer = phrase_buffer.rsplit("\n", 1)
                    phrases.append(phrase)

        if self.tts_enabled and phrase_buffer.strip():
            phrases.append(phrase_buffer)

        conv_log.finish_streaming_response()

        self.chat_history.append({"role": "user", "content": text})
        self.chat_history.append({"role": "assistant", "content": full_response})

        total_tokens = llm_mod.count_conversation_tokens(
            self.models.tokenizer, self.chat_history
        )

        llm_time = time.time() - llm_start
        tts_time: float | None = None

        if self.tts_enabled and phrases:
            self.state = AppState.SYNTHESIZING
            loop = asyncio.get_running_loop()

            def on_play() -> None:
                loop.call_soon_threadsafe(setattr, self, "state", AppState.SPEAKING)

            tts_time = await tts_mod.speak_phrases(
                self.models.tts, phrases, on_first_chunk=on_play
            )

        stats = InteractionStats(
            transcribe_time=transcribe_time,
            llm_time=llm_time,
            tts_time=tts_time,
            tokens=token_count,
            total_conversation_tokens=total_tokens,
            max_tokens=settings.llm_max_tokens,
        )
        metrics_panel.update_metrics(stats)
        self.state = AppState.READY

    def _show_status(self, message: str, timeout: float = 3.0) -> None:
        """Show an ephemeral message in the status bar."""
        status_panel = self.query_one("#status-panel", StatusPanel)
        status_panel.show_ephemeral_message(message, timeout)

    def _show_sticky_status(self, message: str) -> None:
        """Show a status message that persists until explicitly cleared."""
        status_panel = self.query_one("#status-panel", StatusPanel)
        status_panel.show_status_message(message)

    def _reset_esc_pending(self) -> None:
        """Reset the ESC pending state."""
        self._esc_pending = False
        self._esc_timer = None

    def _cancel_processing(self) -> None:
        """Cancel the active processing pipeline."""
        if self._active_worker is not None:
            self._active_worker.cancel()
            self._active_worker = None
        conv_log = self.query_one("#conversation-log", ConversationLog)
        conv_log.finish_streaming_response()
        self.is_processing = False
        self.state = AppState.READY
        self._esc_pending = False
        if self._esc_timer is not None:
            self._esc_timer.stop()
            self._esc_timer = None
        self._show_status("Cancelled")

    async def _run_record_pipeline(self) -> None:
        """Record audio, transcribe, then run LLM pipeline."""
        self.state = AppState.RECORDING

        audio_data = await audio_mod.record()

        if audio_data is None:
            self.state = AppState.READY
            self._show_status("No audio detected.")
            return

        self.state = AppState.TRANSCRIBING
        stt_start = time.time()
        transcribed = await stt_mod.transcribe(self.models.stt, audio_data)
        stt_time = time.time() - stt_start

        if not transcribed:
            self.state = AppState.READY
            self._show_status("Could not transcribe audio.")
            return

        conv_log = self.query_one("#conversation-log", ConversationLog)
        conv_log.add_user_message(transcribed)
        await self._run_llm_pipeline(transcribed, transcribe_time=stt_time)

    async def _run_transcribe(self) -> None:
        """Record extended audio and stream transcription."""
        self.state = AppState.RECORDING
        self._show_sticky_status("Transcribe mode: recording up to 3 min...")

        audio_data = await audio_mod.record_long()

        if audio_data is None:
            self.state = AppState.READY
            self._show_status("No audio detected.")
            return

        self.state = AppState.TRANSCRIBING
        full_text = ""

        async for chunk in stt_mod.transcribe_streaming(self.models.stt, audio_data):
            full_text += chunk

        self.state = AppState.READY
        if full_text.strip():
            text_area = self.query_one("#user-input", ChatInput)
            text_area.clear()
            text_area.insert(full_text.strip())
            self.copy_to_clipboard(full_text.strip())
            self._show_status("Transcribed to input and copied to clipboard.")
        else:
            self._show_status("Could not transcribe audio.")

    def _toggle_tts(self) -> None:
        """Toggle TTS on/off."""
        status_panel = self.query_one("#status-panel", StatusPanel)
        self.tts_enabled = not self.tts_enabled
        status_panel.tts_enabled = self.tts_enabled

    def _open_model_selector(self) -> None:
        """Open the model selector modal."""
        self.push_screen(
            ModelSelector(
                current_stt=settings.stt_model,
                current_llm=settings.llm_model,
                current_tts=settings.tts_model,
            ),
        )

    def on_model_selector_changed(self, event: ModelSelector.Changed) -> None:
        """Handle model selection changes from the modal."""
        self._apply_model_selection(event.selection)

    def _apply_model_selection(self, selection: ModelSelection | None) -> None:
        """Apply model changes from the selector modal."""
        if selection is None:
            return

        reload_stt = selection.stt_model != settings.stt_model
        reload_llm = selection.llm_model != settings.llm_model
        reload_tts = selection.tts_model != settings.tts_model

        if not (reload_stt or reload_llm or reload_tts):
            return

        settings.stt_model = selection.stt_model
        settings.llm_model = selection.llm_model
        settings.tts_model = selection.tts_model

        self.run_worker(
            self._reload_models(
                reload_stt=reload_stt,
                reload_llm=reload_llm,
                reload_tts=reload_tts,
            )
        )

    async def _reload_models(
        self,
        *,
        reload_stt: bool,
        reload_llm: bool,
        reload_tts: bool,
    ) -> None:
        """Reload changed models in the background."""
        loop = asyncio.get_running_loop()
        status_panel = self.query_one("#status-panel", StatusPanel)
        text_area = self.query_one("#user-input", ChatInput)
        text_area.disabled = True
        self.state = AppState.LOADING

        if reload_stt:
            status_panel.show_status_message("Reloading STT...")
            stt = await loop.run_in_executor(None, load_stt, settings.stt_model)
            self.models.stt = stt

        if reload_llm:
            status_panel.show_status_message("Reloading LLM...")
            llm, tokenizer = await loop.run_in_executor(
                None, load_llm, settings.llm_model
            )
            self.models.llm = llm
            self.models.tokenizer = tokenizer

        if reload_tts:
            status_panel.show_status_message("Reloading TTS...")
            tts = await loop.run_in_executor(None, load_tts, settings.tts_model)
            self.models.tts = tts

        status_panel.clear_status_message()
        text_area.disabled = False
        text_area.focus()
        self.state = AppState.READY

        if isinstance(self.screen, ModelSelector):
            self.screen.update_loaded_models(
                stt=settings.stt_model,
                llm=settings.llm_model,
                tts=settings.tts_model,
            )

    def action_clear_conversation(self) -> None:
        """Clear the conversation history."""
        self.chat_history.clear()
        conv_log = self.query_one("#conversation-log", ConversationLog)
        conv_log.clear()

        metrics_panel = self.query_one("#metrics-panel", MetricsPanel)
        metrics_panel.clear_metrics()



"""Custom Textual widgets for Voxing TUI."""

from voxing.widgets.chat_input import ChatInput
from voxing.widgets.conversation import ConversationLog
from voxing.widgets.metrics_panel import MetricsPanel
from voxing.widgets.model_selector import ModelSelection, ModelSelector
from voxing.widgets.status_panel import StatusPanel

__all__ = [
    "ChatInput",
    "ConversationLog",
    "MetricsPanel",
    "ModelSelection",
    "ModelSelector",
    "StatusPanel",
]



"""Audio recording and playback via sounddevice."""

import asyncio
from collections.abc import Callable, Iterable
from functools import partial

import numpy as np
import sounddevice as sd

from voxing.config import settings


def _rms(audio: np.ndarray) -> float:
    """Calculate root mean square of audio signal."""
    return float(np.sqrt(np.mean(audio**2)))


def _normalize_audio(audio: np.ndarray) -> np.ndarray:
    """Normalize audio to prevent clipping."""
    audio = audio.flatten().astype(np.float32)
    max_val = np.abs(audio).max()
    if max_val > 0:
        audio = audio / max_val * 0.9
    return audio.reshape(-1, 1)


def _record_blocking(
    silence_duration: float = 0.7,
    max_duration: float = 10.0,
) -> np.ndarray | None:
    """Record audio until silence is detected."""
    chunk_duration = 0.1
    chunk_samples = int(settings.audio_sample_rate * chunk_duration)
    silence_chunks_needed = int(silence_duration / chunk_duration)
    max_chunks = int(max_duration / chunk_duration)

    audio_chunks: list[np.ndarray] = []
    silence_count = 0
    recording = False

    with sd.InputStream(
        samplerate=settings.audio_sample_rate, channels=1, dtype=np.float32
    ) as stream:
        while len(audio_chunks) < max_chunks:
            chunk, _ = stream.read(chunk_samples)
            has_voice = _rms(chunk) > settings.silence_threshold

            if not recording:
                if has_voice:
                    recording = True
                    audio_chunks.append(chunk)
                    silence_count = 0
                continue

            audio_chunks.append(chunk)

            if has_voice:
                silence_count = 0
            else:
                silence_count += 1
                if silence_count >= silence_chunks_needed:
                    break

    if not audio_chunks:
        return None
    return np.concatenate(audio_chunks, axis=0).flatten()


def _play_stream(
    chunks: Iterable[np.ndarray],
    sample_rate: int,
    on_first_chunk: Callable[[], None] | None = None,
) -> None:
    """Play an iterable of audio chunks through the speakers."""
    first = True
    with sd.OutputStream(
        samplerate=sample_rate,
        channels=1,
        dtype=np.float32,
        blocksize=4096,
    ) as stream:
        for chunk in chunks:
            audio = np.array(chunk, dtype=np.float32)
            if first and on_first_chunk is not None:
                on_first_chunk()
                first = False
            stream.write(_normalize_audio(audio))


async def record() -> np.ndarray | None:
    """Record a short voice command."""
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(
        None, partial(_record_blocking, silence_duration=0.7, max_duration=10.0)
    )


async def record_long() -> np.ndarray | None:
    """Record extended speech for transcription."""
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(
        None, partial(_record_blocking, silence_duration=3.0, max_duration=180.0)
    )


async def play_stream(
    chunks: Iterable[np.ndarray],
    sample_rate: int,
    on_first_chunk: Callable[[], None] | None = None,
) -> None:
    """Play audio chunks asynchronously."""
    loop = asyncio.get_running_loop()
    await loop.run_in_executor(
        None, partial(_play_stream, chunks, sample_rate, on_first_chunk)
    )



"""Speech-to-text wrapper."""

import asyncio
from collections.abc import AsyncGenerator
from functools import partial

import mlx.core as mx
import mlx.nn as nn
import numpy as np

from voxing.models._streaming import sync_to_async_iter


async def transcribe(stt_model: nn.Module, audio: np.ndarray) -> str:
    """Transcribe audio using STT model."""
    audio_mx = mx.array(audio)
    loop = asyncio.get_running_loop()
    result = await loop.run_in_executor(None, partial(stt_model.generate, audio_mx))
    return result.text.strip()


async def transcribe_streaming(
    stt_model: nn.Module, audio: np.ndarray
) -> AsyncGenerator[str]:
    """Transcribe audio with streaming output, yielding text chunks."""
    audio_mx = mx.array(audio)
    async for chunk in sync_to_async_iter(stt_model.stream_generate, audio_mx):
        if chunk.text:
            yield chunk.text



"""Tokyo Night color scheme for Voxing TUI."""

from textual.design import ColorSystem

# Tokyo Night Night variant - Official color palette
# Source: https://github.com/folke/tokyonight.nvim/blob/main/extras/ghostty/tokyonight_night

# Standard palette (0-8)
PALETTE_0 = "#15161e"  # Darker surface
PALETTE_1 = "#f7768e"  # Red
PALETTE_2 = "#9ece6a"  # Green
PALETTE_3 = "#e0af68"  # Yellow
PALETTE_4 = "#7aa2f7"  # Blue
PALETTE_5 = "#bb9af7"  # Purple/Magenta
PALETTE_6 = "#7dcfff"  # Cyan
PALETTE_7 = "#a9b1d6"  # Light text
PALETTE_8 = "#414868"  # Panel/border

# Bright variants (9-15)
BRIGHT_RED = "#ff899d"  # palette 9
BRIGHT_GREEN = "#9fe044"  # palette 10
BRIGHT_YELLOW = "#faba4a"  # palette 11
BRIGHT_BLUE = "#8db0ff"  # palette 12
BRIGHT_PURPLE = "#c7a9ff"  # palette 13
BRIGHT_CYAN = "#a4daff"  # palette 14
BRIGHT_WHITE = "#c0caf5"  # palette 15

# Official Tokyo Night special colors
BACKGROUND = "#1a1b26"  # Dark background
FOREGROUND = "#c0caf5"  # Main text (same as BRIGHT_WHITE)
SELECTION_BG = "#283457"  # Selection background

# Textual ColorSystem (9 semantic colors for Textual framework)
TOKYO_NIGHT = ColorSystem(
    primary=PALETTE_4,  # Blue
    secondary=PALETTE_5,  # Purple/Magenta
    accent=PALETTE_6,  # Cyan
    background=BACKGROUND,
    surface=PALETTE_0,
    panel=PALETTE_8,
    success=PALETTE_2,  # Green
    warning=PALETTE_3,  # Yellow
    error=PALETTE_1,  # Red
    dark=True,
)



"""Modal screen for switching models."""

from dataclasses import dataclass
from typing import Literal

from rich.text import Text
from textual.app import ComposeResult
from textual.containers import Vertical, VerticalScroll
from textual.message import Message
from textual.screen import ModalScreen
from textual.widgets import Label, Static

from voxing.config import LLM_MODELS, STT_MODELS, TTS_MODELS
from voxing.themes import PALETTE_2, PALETTE_4, PALETTE_8

ModelType = Literal["stt", "llm", "tts"]


@dataclass(slots=True)
class ModelSelection:
    """Result of model selection dialog."""

    stt_model: str
    llm_model: str
    tts_model: str


class ModelItem(Static, can_focus=True):
    """Single model item in the list."""

    def __init__(self, model_name: str, is_loaded: bool) -> None:
        super().__init__()
        self.model_name = model_name
        self.is_loaded = is_loaded

    def render(self) -> Text:
        """Render model item with appropriate styling."""
        text = Text()

        if self.has_focus:
            text.append("▸ ", style=PALETTE_4)
        else:
            text.append("  ")

        style = f"bold {PALETTE_4}" if self.has_focus else PALETTE_8
        text.append(self.model_name, style=style)

        if self.is_loaded:
            text.append(" ✓", style=PALETTE_2)

        return text

    def update_loaded(self, is_loaded: bool) -> None:
        """Update loaded state and re-render."""
        self.is_loaded = is_loaded
        self.refresh()

    def on_focus(self) -> None:
        """Called when item gains focus - trigger re-render."""
        self.refresh()

    def on_blur(self) -> None:
        """Called when item loses focus - trigger re-render."""
        self.refresh()


class ModelSelector(ModalScreen[ModelSelection | None]):
    """Modal screen for selecting models across STT, LLM, and TTS."""

    class Changed(Message):
        """Posted when model selection changes."""

        def __init__(self, selection: ModelSelection) -> None:
            super().__init__()
            self.selection = selection

    BINDINGS = [
        ("escape", "cancel", "Close"),
        ("enter", "select", "Select"),
        ("up,k", "cursor_up", "Up"),
        ("down,j", "cursor_down", "Down"),
    ]

    def __init__(
        self,
        current_stt: str,
        current_llm: str,
        current_tts: str,
    ) -> None:
        """Initialize with current model selections."""
        super().__init__()
        self._current_stt = current_stt
        self._current_llm = current_llm
        self._current_tts = current_tts
        self._all_models: list[tuple[ModelType, str]] = []

    def compose(self) -> ComposeResult:
        """Compose the modal layout."""
        with Vertical(id="model-dialog"), VerticalScroll(id="model-list"):
            for label, models, current in [
                ("STT", STT_MODELS, self._current_stt),
                ("LLM", LLM_MODELS, self._current_llm),
                ("TTS", TTS_MODELS, self._current_tts),
            ]:
                yield Label(label, classes="model-header")
                model_type = label.lower()
                for model in models:
                    self._all_models.append((model_type, model))  # type: ignore[arg-type]
                    yield ModelItem(model, is_loaded=model == current)

    def on_mount(self) -> None:
        """Set initial focus on first model item."""
        items = self._get_model_items()
        if items:
            items[0].focus()

    def _get_model_items(self) -> list[ModelItem]:
        """Get all ModelItem widgets."""
        return list(self.query(ModelItem))

    def _get_focused_item_index(self) -> int:
        """Get index of currently focused ModelItem."""
        items = self._get_model_items()
        for i, item in enumerate(items):
            if item.has_focus:
                return i
        return 0

    def action_cursor_up(self) -> None:
        """Move focus to previous item."""
        items = self._get_model_items()
        if not items:
            return

        current = self._get_focused_item_index()
        previous = (current - 1) % len(items)
        items[previous].focus()

    def action_cursor_down(self) -> None:
        """Move focus to next item."""
        items = self._get_model_items()
        if not items:
            return

        current = self._get_focused_item_index()
        next_idx = (current + 1) % len(items)
        items[next_idx].focus()

    def action_select(self) -> None:
        """Select currently focused model and apply changes."""
        current = self._get_focused_item_index()

        if current >= len(self._all_models):
            return

        model_type, model_name = self._all_models[current]

        selection = ModelSelection(
            stt_model=model_name if model_type == "stt" else self._current_stt,
            llm_model=model_name if model_type == "llm" else self._current_llm,
            tts_model=model_name if model_type == "tts" else self._current_tts,
        )
        self.post_message(self.Changed(selection))

    def _update_all_items(self) -> None:
        """Update all model items to reflect current loaded state."""
        items = self._get_model_items()
        for i, item in enumerate(items):
            if i >= len(self._all_models):
                continue

            model_type, model_name = self._all_models[i]

            is_loaded = (
                (model_type == "stt" and model_name == self._current_stt)
                or (model_type == "llm" and model_name == self._current_llm)
                or (model_type == "tts" and model_name == self._current_tts)
            )

            item.update_loaded(is_loaded)

    def update_loaded_models(self, stt: str, llm: str, tts: str) -> None:
        """Update which models are marked as loaded."""
        self._current_stt = stt
        self._current_llm = llm
        self._current_tts = tts
        self._update_all_items()

    def action_cancel(self) -> None:
        """Cancel and close the modal."""
        self.dismiss(None)



from textual.binding import Binding
from textual.message import Message
from textual.widgets import TextArea


class ChatInput(TextArea):
    """TextArea subclass where Enter submits and Ctrl+J inserts a newline."""

    BINDINGS = [
        Binding("enter", "submit", "Submit", priority=True),
        Binding("Shift+enter,ctrl+j", "newline", "Newline"),
    ]

    class Submitted(Message):
        """Posted when the user presses Enter to submit input."""

        def __init__(self, chat_input: "ChatInput", value: str) -> None:
            super().__init__()
            self.chat_input = chat_input
            self.value = value

    def action_submit(self) -> None:
        """Submit the current text."""
        self.post_message(self.Submitted(self, self.text))

    def action_newline(self) -> None:
        """Insert a newline character."""
        self.insert("\n")



"""Conversation log widget for displaying chat history."""

from rich.text import Text
from textual.geometry import Size
from textual.widgets import RichLog

from voxing.themes import FOREGROUND, PALETTE_1, PALETTE_3, PALETTE_5, PALETTE_6


class ConversationLog(RichLog):
    """Widget for displaying conversation history with rich formatting."""

    _is_streaming: bool = False
    _streaming_text: str = ""
    _streaming_line_count: int = 0

    def add_user_message(self, text: str) -> None:
        """Add a user message to the conversation log."""
        message = Text()
        message.append("You: ", style=f"bold {PALETTE_6}")
        message.append(text, style=FOREGROUND)
        self.write(message)

    def add_system_message(self, text: str, style: str = PALETTE_3) -> None:
        """Add a system message to the conversation log."""
        message = Text()
        message.append(f"[{text}]", style=style)
        self.write(message)

    def start_streaming_response(self) -> None:
        """Start a new streaming assistant response."""
        self._streaming_text = ""
        self._is_streaming = True
        self._streaming_line_count = 0

    def _pop_streaming_lines(self) -> None:
        """Remove all lines belonging to the current streaming message."""
        for _ in range(self._streaming_line_count):
            if self.lines:
                self.lines.pop()
        self._line_cache.clear()
        self.virtual_size = Size(self.virtual_size.width, len(self.lines))

    def update_streaming_response(self, token: str) -> None:
        """Update the streaming response with a new token."""
        if not self._is_streaming:
            self.start_streaming_response()

        self._streaming_text += token

        message = Text()
        message.append("Assistant: ", style=f"bold {PALETTE_5}")
        message.append(self._streaming_text + " \u2588", style=FOREGROUND)

        self._pop_streaming_lines()
        lines_before = len(self.lines)
        self.write(message)
        self._streaming_line_count = len(self.lines) - lines_before

    def finish_streaming_response(self) -> None:
        """Finish the streaming response."""
        if not self._streaming_text or not self._is_streaming:
            return

        self._pop_streaming_lines()

        message = Text()
        message.append("Assistant: ", style=f"bold {PALETTE_5}")
        message.append(self._streaming_text, style=FOREGROUND)
        self.write(message)

        self._is_streaming = False
        self._streaming_text = ""
        self._streaming_line_count = 0

    def add_error(self, error_text: str) -> None:
        """Add an error message."""
        self.add_system_message(f"Error: {error_text}", style=f"bold {PALETTE_1}")



"""Kokoro text-to-speech wrapper."""

import asyncio
import time
from collections.abc import Callable, Iterator, Sequence
from functools import partial

import mlx.nn as nn
import numpy as np
from langdetect import detect

from voxing.audio import play_stream
from voxing.config import settings

KOKORO_VOICES: dict[str, tuple[str, str]] = {
    "en": ("af_heart", "a"),
    "es": ("ef_dora", "e"),
    "fr": ("ff_siwis", "f"),
    "hi": ("hf_alpha", "h"),
    "it": ("if_sara", "i"),
    "pt": ("pf_dora", "p"),
}

DEFAULT_VOICE = ("af_heart", "a")


def _detect_voice(text: str) -> tuple[str, str]:
    """Detect language and return the matching Kokoro voice and lang code."""
    detected = detect(text)
    return KOKORO_VOICES.get(detected, DEFAULT_VOICE)


def _synthesize(tts_model: nn.Module, text: str) -> Iterator[np.ndarray]:
    """Yield audio chunks from the TTS model."""
    if not text.strip():
        return
    voice, lang_code = _detect_voice(text)
    for result in tts_model.generate(
        text=text, voice=voice, lang_code=lang_code, speed=1.0
    ):
        yield np.array(result.audio, dtype=np.float32)


def _synthesize_phrases(
    tts_model: nn.Module,
    phrases: Sequence[str],
) -> Iterator[np.ndarray]:
    """Yield audio chunks from a sequence of phrases."""
    for phrase in phrases:
        yield from _synthesize(tts_model, phrase)


async def speak_phrases(
    tts_model: nn.Module,
    phrases: Sequence[str],
    on_first_chunk: Callable[[], None] | None = None,
) -> float:
    """Synthesize and play a sequence of phrases. Returns synthesis time."""
    loop = asyncio.get_running_loop()

    synth_start = time.time()
    chunks = await loop.run_in_executor(
        None, partial(list, _synthesize_phrases(tts_model, phrases))
    )
    synth_time = time.time() - synth_start

    await play_stream(chunks, settings.tts_sample_rate, on_first_chunk)
    return synth_time



"""Model loading and initialization."""

import contextlib
import logging
import os
import warnings
from collections.abc import Iterator
from dataclasses import dataclass
from pathlib import Path

import mlx.nn as nn
from mlx_audio.stt import load_model as load_stt_model
from mlx_audio.tts.utils import load_model as load_tts_model
from mlx_lm import load
from mlx_lm.tokenizer_utils import TokenizerWrapper

from voxing.config import settings

for _name in ("mlx", "mlx_lm", "mlx_audio"):
    logging.getLogger(_name).setLevel(logging.ERROR)
warnings.filterwarnings("ignore", message=".*deprecated.*")


@contextlib.contextmanager
def suppress_output() -> Iterator[None]:
    """Redirect stdout/stderr to devnull at both Python and OS fd level."""
    with Path(os.devnull).open("w") as devnull:
        saved_fds = os.dup(1), os.dup(2)
        os.dup2(devnull.fileno(), 1)
        os.dup2(devnull.fileno(), 2)
        try:
            with (
                contextlib.redirect_stdout(devnull),
                contextlib.redirect_stderr(devnull),
            ):
                yield
        finally:
            os.dup2(saved_fds[0], 1)
            os.dup2(saved_fds[1], 2)
            os.close(saved_fds[0])
            os.close(saved_fds[1])


@dataclass(slots=True)
class Models:
    """Container for all loaded models."""

    stt: nn.Module
    llm: nn.Module
    tts: nn.Module
    tokenizer: TokenizerWrapper


def load_stt(model_name: str | None = None) -> nn.Module:
    """Load the speech-to-text model."""
    with suppress_output():
        return load_stt_model(model_name or settings.stt_model)


def load_llm(model_name: str | None = None) -> tuple[nn.Module, TokenizerWrapper]:
    """Load the LLM model and tokenizer."""
    with suppress_output():
        model, tokenizer = load(model_name or settings.llm_model)  # ty:ignore[invalid-assignment]
    return model, tokenizer


def load_tts(model_name: str | None = None) -> nn.Module:
    """Load the text-to-speech model."""
    with suppress_output():
        return load_tts_model(model_name or settings.tts_model)  # ty:ignore[invalid-argument-type]



"""Application settings via pydantic-settings."""

from pydantic_settings import BaseSettings, SettingsConfigDict

DEFAULT_STT_MODEL = "mlx-community/parakeet-tdt-0.6b-v3"
DEFAULT_LLM_MODEL = "LiquidAI/LFM2.5-1.2B-Instruct-MLX-8bit"
DEFAULT_TTS_MODEL = "mlx-community/Kokoro-82M-bf16"

STT_MODELS: tuple[str, ...] = (
    DEFAULT_STT_MODEL,
    "mlx-community/whisper-large-v3-turbo-asr-fp16",
)

LLM_MODELS: tuple[str, ...] = (
    DEFAULT_LLM_MODEL,
    "mlx-community/Nanbeige4.1-3B-8bit",
    "Qwen/Qwen3-0.6B-MLX-4bit",
)

TTS_MODELS: tuple[str, ...] = (DEFAULT_TTS_MODEL,)


class Settings(BaseSettings):
    """Voxing TUI configuration, overridable via VOX_ env vars or .env file."""

    model_config = SettingsConfigDict(env_prefix="VOX_")

    stt_model: str = DEFAULT_STT_MODEL
    llm_model: str = DEFAULT_LLM_MODEL
    tts_model: str = DEFAULT_TTS_MODEL
    system_prompt: str = "You are a helpful assistant"
    llm_max_tokens: int = 4096
    audio_sample_rate: int = 16000
    silence_threshold: float = 0.01
    tts_sample_rate: int = 24000


settings = Settings()



"""Async bridge for synchronous iterators."""

import asyncio
from collections.abc import AsyncGenerator, Callable, Iterator


async def sync_to_async_iter[T](
    fn: Callable[..., Iterator[T]],
    *args: object,
    **kwargs: object,
) -> AsyncGenerator[T]:
    """Bridge a synchronous iterator into an async generator via a thread executor."""
    sentinel = object()
    iterator = await asyncio.to_thread(fn, *args, **kwargs)
    while (item := await asyncio.to_thread(next, iterator, sentinel)) is not sentinel:
        yield item



"""Metrics panel widget for displaying performance statistics."""

from rich.text import Text
from textual.widgets import Static

from voxing.state import InteractionStats
from voxing.themes import FOREGROUND


class MetricsPanel(Static):
    """Widget for displaying performance metrics."""

    current_stats: InteractionStats | None = None

    def update_metrics(self, stats: InteractionStats) -> None:
        """Update the displayed metrics."""
        self.current_stats = stats
        self.update_display()

    def update_display(self) -> None:
        """Update the panel display with current metrics."""
        content = Text()

        if self.current_stats is None:
            self.update(content)
            return
        stats = self.current_stats

        if stats.transcribe_time is not None:
            content.append("STT: ", style="dim")
            content.append(f"{stats.transcribe_time:.2f}s", style=FOREGROUND)
            content.append("  ")

        content.append("LLM: ", style="dim")
        content.append(f"{stats.llm_time:.1f}s", style=FOREGROUND)
        content.append("  Tokens: ", style="dim")
        content.append(f"{stats.tokens_per_sec:.0f}/s", style=FOREGROUND)

        if stats.tts_time is not None:
            content.append("  TTS: ", style="dim")
            content.append(f"{stats.tts_time:.2f}s", style=FOREGROUND)

        if stats.max_tokens > 0:
            content.append("  Tokens: ", style="dim")
            content.append(
                f"{stats.total_conversation_tokens}/{stats.max_tokens}",
                style=FOREGROUND,
            )

        self.update(content)

    def clear_metrics(self) -> None:
        """Clear the current metrics."""
        self.current_stats = None
        self.update_display()

    def on_mount(self) -> None:
        """Called when widget is mounted."""
        self.update_display()



"""Entry point for Voxinging TUI application."""

from voxing.app import VoxingApp


def main() -> None:
    """Run the Voxing TUI application."""
    app = VoxingApp()
    app.run()


if __name__ == "__main__":
    main()



"""State management data structures for Voxing TUI."""

from dataclasses import dataclass
from enum import Enum


class AppState(Enum):
    """Application state machine states."""

    LOADING = "loading"
    READY = "ready"
    RECORDING = "recording"
    TRANSCRIBING = "transcribing"
    THINKING = "thinking"
    SYNTHESIZING = "synthesizing"
    SPEAKING = "speaking"

    def __str__(self) -> str:
        """Return human-readable state name."""
        return self.value.capitalize()


@dataclass
class InteractionStats:
    """Performance metrics for a single interaction."""

    transcribe_time: float | None = None
    llm_time: float = 0.0
    tts_time: float | None = None
    tokens: int = 0
    tokens_per_sec: float = 0.0
    total_conversation_tokens: int = 0
    max_tokens: int = 0

    def __post_init__(self) -> None:
        """Calculate derived metrics."""
        if self.llm_time > 0 and self.tokens > 0:
            self.tokens_per_sec = self.tokens / self.llm_time



"""LFM2.5 streaming LLM wrapper."""

from collections.abc import AsyncGenerator

import mlx.nn as nn
from mlx_lm import stream_generate
from mlx_lm.sample_utils import make_logits_processors, make_sampler
from mlx_lm.tokenizer_utils import TokenizerWrapper

from voxing.config import settings
from voxing.models._streaming import sync_to_async_iter

type ChatMessage = dict[str, str]


def count_tokens(tokenizer: TokenizerWrapper, text: str) -> int:
    """Count the number of tokens in a text string."""
    return len(tokenizer.encode(text))


def count_conversation_tokens(
    tokenizer: TokenizerWrapper,
    history: list[ChatMessage],
    system_prompt: str = settings.system_prompt,
) -> int:
    """Count total tokens in conversation history including system prompt."""
    messages: list[ChatMessage] = [{"role": "system", "content": system_prompt}]
    messages.extend(history)
    prompt = tokenizer.apply_chat_template(
        messages, add_generation_prompt=False, tokenize=False
    )
    return count_tokens(tokenizer, prompt)


def _format_prompt(
    tokenizer: TokenizerWrapper,
    user_input: str,
    history: list[ChatMessage] | None = None,
) -> str:
    """Format user input into a chat prompt with optional history."""
    messages: list[ChatMessage] = [
        {"role": "system", "content": settings.system_prompt}
    ]
    if history:
        messages.extend(history)
    messages.append({"role": "user", "content": user_input})
    return tokenizer.apply_chat_template(
        messages, add_generation_prompt=True, tokenize=False
    )


async def generate_streaming(
    model: nn.Module,
    tokenizer: TokenizerWrapper,
    user_input: str,
    max_tokens: int = settings.llm_max_tokens,
    history: list[ChatMessage] | None = None,
) -> AsyncGenerator[str]:
    """Async generator that yields tokens as they are generated."""
    prompt = _format_prompt(tokenizer, user_input, history)
    sampler = make_sampler(temp=0.1, top_k=50, top_p=0.1)
    logits_processors = make_logits_processors(repetition_penalty=1.05)

    async for chunk in sync_to_async_iter(
        stream_generate,
        model=model,
        tokenizer=tokenizer,
        prompt=prompt,
        max_tokens=max_tokens,
        sampler=sampler,
        logits_processors=logits_processors,
    ):
        yield chunk.text
